@@ -2,22 +2,24 @@
  *
  * rewriteManip.c
  *
- * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2005, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql/src/backend/rewrite/rewriteManip.c,v 1.80 2003/10/20 20:01:59 tgl Exp $
+ *	  $PostgreSQL: pgsql/src/backend/rewrite/rewriteManip.c,v 1.89.4.1 2006/01/06 20:11:24 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
 #include "postgres.h"
 
+#include "catalog/pg_type.h"
 #include "nodes/makefuncs.h"
 #include "optimizer/clauses.h"
 #include "optimizer/tlist.h"
 #include "parser/parsetree.h"
-#include "parser/parse_clause.h"
+#include "parser/parse_coerce.h"
+#include "parser/parse_relation.h"
 #include "rewrite/rewriteManip.h"
 #include "utils/lsyscache.h"
 
@@ -216,7 +218,6 @@ OffsetVarNodes(Node *node, int offset, int sublevels_up)
 	if (node && IsA(node, Query))
 	{
 		Query	   *qry = (Query *) node;
-		List	   *l;
 
 		/*
 		 * If we are starting at a Query, and sublevels_up is zero, then
@@ -227,10 +228,12 @@ OffsetVarNodes(Node *node, int offset, int sublevels_up)
 		 */
 		if (sublevels_up == 0)
 		{
+			ListCell   *l;
+
 			if (qry->resultRelation)
 				qry->resultRelation += offset;
 			foreach(l, qry->rowMarks)
-				lfirsti(l) += offset;
+				lfirst_int(l) += offset;
 		}
 		query_tree_walker(qry, OffsetVarNodes_walker,
 						  (void *) &context, 0);
@@ -356,7 +359,6 @@ ChangeVarNodes(Node *node, int rt_index, int new_index, int sublevels_up)
 	if (node && IsA(node, Query))
 	{
 		Query	   *qry = (Query *) node;
-		List	   *l;
 
 		/*
 		 * If we are starting at a Query, and sublevels_up is zero, then
@@ -367,12 +369,14 @@ ChangeVarNodes(Node *node, int rt_index, int new_index, int sublevels_up)
 		 */
 		if (sublevels_up == 0)
 		{
+			ListCell   *l;
+
 			if (qry->resultRelation == rt_index)
 				qry->resultRelation = new_index;
 			foreach(l, qry->rowMarks)
 			{
-				if (lfirsti(l) == rt_index)
-					lfirsti(l) = new_index;
+				if (lfirst_int(l) == rt_index)
+					lfirst_int(l) = new_index;
 			}
 		}
 		query_tree_walker(qry, ChangeVarNodes_walker,
@@ -669,23 +673,23 @@ getInsertSelectQuery(Query *parsetree, Query ***subquery_ptr)
 	 * query.  If they're not there, it must be an INSERT/SELECT in which
 	 * they've been pushed down to the SELECT.
 	 */
-	if (length(parsetree->rtable) >= 2 &&
+	if (list_length(parsetree->rtable) >= 2 &&
 	 strcmp(rt_fetch(PRS2_OLD_VARNO, parsetree->rtable)->eref->aliasname,
 			"*OLD*") == 0 &&
 	 strcmp(rt_fetch(PRS2_NEW_VARNO, parsetree->rtable)->eref->aliasname,
 			"*NEW*") == 0)
 		return parsetree;
 	Assert(parsetree->jointree && IsA(parsetree->jointree, FromExpr));
-	if (length(parsetree->jointree->fromlist) != 1)
+	if (list_length(parsetree->jointree->fromlist) != 1)
 		elog(ERROR, "expected to find SELECT subquery");
-	rtr = (RangeTblRef *) lfirst(parsetree->jointree->fromlist);
+	rtr = (RangeTblRef *) linitial(parsetree->jointree->fromlist);
 	Assert(IsA(rtr, RangeTblRef));
 	selectrte = rt_fetch(rtr->rtindex, parsetree->rtable);
 	selectquery = selectrte->subquery;
 	if (!(selectquery && IsA(selectquery, Query) &&
 		  selectquery->commandType == CMD_SELECT))
 		elog(ERROR, "expected to find SELECT subquery");
-	if (length(selectquery->rtable) >= 2 &&
+	if (list_length(selectquery->rtable) >= 2 &&
 	strcmp(rt_fetch(PRS2_OLD_VARNO, selectquery->rtable)->eref->aliasname,
 		   "*OLD*") == 0 &&
 	strcmp(rt_fetch(PRS2_NEW_VARNO, selectquery->rtable)->eref->aliasname,
@@ -850,6 +854,11 @@ AddInvertedQual(Query *parsetree, Node *qual)
  * If not, we either change the unmatched Var's varno to update_varno
  * (when event == CMD_UPDATE) or replace it with a constant NULL.
  *
+ * The caller must also provide target_rtable, the rangetable containing
+ * the target relation (which must be described by the target_varno'th
+ * RTE in that list).  This is needed to handle whole-row Vars referencing
+ * the target.	We expand such Vars into RowExpr constructs.
+ *
  * Note: the business with inserted_sublink is needed to update hasSubLinks
  * in subqueries when the replacement adds a subquery inside a subquery.
  * Messy, isn't it?  We do not need to do similar pushups for hasAggs,
@@ -861,12 +870,58 @@ typedef struct
 {
 	int			target_varno;
 	int			sublevels_up;
+	List	   *target_rtable;
 	List	   *targetlist;
 	int			event;
 	int			update_varno;
 	bool		inserted_sublink;
 } ResolveNew_context;
 
+static Node *
+resolve_one_var(Var *var, ResolveNew_context *context)
+{
+	TargetEntry *tle;
+
+	tle = get_tle_by_resno(context->targetlist, var->varattno);
+
+	if (tle == NULL)
+	{
+		/* Failed to find column in insert/update tlist */
+		if (context->event == CMD_UPDATE)
+		{
+			/* For update, just change unmatched var's varno */
+			var = (Var *) copyObject(var);
+			var->varno = context->update_varno;
+			var->varnoold = context->update_varno;
+			return (Node *) var;
+		}
+		else
+		{
+			/* Otherwise replace unmatched var with a null */
+			/* need coerce_to_domain in case of NOT NULL domain constraint */
+			return coerce_to_domain((Node *) makeNullConst(var->vartype),
+									InvalidOid,
+									var->vartype,
+									COERCE_IMPLICIT_CAST,
+									false,
+									false);
+		}
+	}
+	else
+	{
+		/* Make a copy of the tlist item to return */
+		Node	   *n = copyObject(tle->expr);
+
+		/* Adjust varlevelsup if tlist item is from higher query */
+		if (var->varlevelsup > 0)
+			IncrementVarSublevelsUp(n, var->varlevelsup, 0);
+		/* Report it if we are adding a sublink to query */
+		if (!context->inserted_sublink)
+			context->inserted_sublink = checkExprHasSubLink(n);
+		return n;
+	}
+}
+
 static Node *
 ResolveNew_mutator(Node *node, ResolveNew_context *context)
 {
@@ -881,45 +936,34 @@ ResolveNew_mutator(Node *node, ResolveNew_context *context)
 		if (this_varno == context->target_varno &&
 			this_varlevelsup == context->sublevels_up)
 		{
-			TargetEntry *tle;
-
-			/* band-aid: don't do the wrong thing with a whole-tuple Var */
 			if (var->varattno == InvalidAttrNumber)
-				ereport(ERROR,
-						(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
-						 errmsg("cannot handle whole-row reference")));
-
-			tle = get_tle_by_resno(context->targetlist, var->varattno);
-
-			if (tle == NULL)
 			{
-				if (context->event == CMD_UPDATE)
-				{
-					/* For update, just change unmatched var's varno */
-					var = (Var *) copyObject(node);
-					var->varno = context->update_varno;
-					var->varnoold = context->update_varno;
-					return (Node *) var;
-				}
-				else
-				{
-					/* Otherwise replace unmatched var with a null */
-					return (Node *) makeNullConst(var->vartype);
-				}
-			}
-			else
-			{
-				/* Make a copy of the tlist item to return */
-				Node	   *n = copyObject(tle->expr);
+				/* Must expand whole-tuple reference into RowExpr */
+				RowExpr    *rowexpr;
+				List	   *fields;
 
-				/* Adjust varlevelsup if tlist item is from higher query */
-				if (this_varlevelsup > 0)
-					IncrementVarSublevelsUp(n, this_varlevelsup, 0);
-				/* Report it if we are adding a sublink to query */
-				if (!context->inserted_sublink)
-					context->inserted_sublink = checkExprHasSubLink(n);
-				return n;
+				/*
+				 * If generating an expansion for a var of a named rowtype
+				 * (ie, this is a plain relation RTE), then we must
+				 * include dummy items for dropped columns.  If the var is
+				 * RECORD (ie, this is a JOIN), then omit dropped columns.
+				 */
+				expandRTE(context->target_rtable, this_varno, this_varlevelsup,
+						  (var->vartype != RECORDOID),
+						  NULL, &fields);
+				/* Adjust the generated per-field Vars... */
+				fields = (List *) ResolveNew_mutator((Node *) fields,
+													 context);
+				rowexpr = makeNode(RowExpr);
+				rowexpr->args = fields;
+				rowexpr->row_typeid = var->vartype;
+				rowexpr->row_format = COERCE_IMPLICIT_CAST;
+
+				return (Node *) rowexpr;
 			}
+
+			/* Normal case for scalar variable */
+			return resolve_one_var(var, context);
 		}
 		/* otherwise fall through to copy the var normally */
 	}
@@ -948,12 +992,14 @@ ResolveNew_mutator(Node *node, ResolveNew_context *context)
 
 Node *
 ResolveNew(Node *node, int target_varno, int sublevels_up,
+		   List *target_rtable,
 		   List *targetlist, int event, int update_varno)
 {
 	ResolveNew_context context;
 
 	context.target_varno = target_varno;
 	context.sublevels_up = sublevels_up;
+	context.target_rtable = target_rtable;
 	context.targetlist = targetlist;
 	context.event = event;
 	context.update_varno = update_varno;
